/////////////////////////////////////////////////////////////////////////////
// Name:      MainWin.cpp
// Author:    Alex Thuering
// Created:   7.07.2012
// Copyright: (c) Alex Thuering
// Licence:   GPL
/////////////////////////////////////////////////////////////////////////////
#include "MainWin.h"

//(*InternalHeaders(MainWin)
#include <wx/intl.h>
#include <wx/string.h>
//*)
#include <wx/toolbar.h>
#include <wx/progdlg.h>
#include <wx/file.h>
#include <wx/process.h>
#include <wx/txtstrm.h>
#include <wx/regex.h>
#include <wx/artprov.h>
#include <wx/aboutdlg.h>
#include "Languages.h"
#include "Config.h"
#include "Version.h"
#include "utils.h"
#include "../resources/mp4joiner.png.h"
#include "../resources/add.png.h"
#include "../resources/remove.png.h"
#include "../resources/run.png.h"
#include "../resources/preferences.png.h"

class MP4Process: public wxProcess {
public:
	MP4Process(wxProgressDialog* parent, vector<MediaFile*> mediaFiles) : wxProcess(parent), files(mediaFiles) {
		progDlg = parent;
		terminated = false;
		step = 0;
		pattern.Compile(wxT(".*\\(([0-9]+)/100\\).*"));
	}
	
	virtual void OnTerminate(int pid, int status) {
		while (HasInput())
			; // read the rest of the output
		progDlg->Update((files.size()+1)*100, status == 0 ? _("Successful") : _("Failed"));
		terminated = true;
	}
	
	bool IsTerminated() {
		return terminated;
	}
	
	bool HasInput() {
		bool hasInput = false;
		
		if (IsInputAvailable()) {
			DoGetFromStream(*GetInputStream(), lineOut, false);
			hasInput = true;
		}

		if (IsErrorAvailable()) {
			DoGetFromStream(*GetErrorStream(), lineErr, true);
			hasInput = true;
		}
		
		return hasInput;
	}
    
	/**
	 * Reads data from given stream
	 */
    void DoGetFromStream(wxInputStream& in, wxString& line, bool errorStream) {
		wxTextInputStream tis(in);
		while (in.CanRead()) {
			wxYieldIfNeeded();
			while (in.CanRead()) {
				wxChar c = tis.GetChar();
				if (c == wxT('\n') || c == wxT('\r')) {
					if (line.length()) {
						ProcessOutput(line, errorStream);
					}
					line = wxT("");
					break;
				}
				line += c;
			}
		}
	}
    
    void ProcessOutput(const wxString& line, bool errorStream) {
    	if (line.StartsWith(wxT("Error")) || errorStream) {
			wxLogError(line);
		} else if (line.StartsWith(wxT("Appending file"))) {
			progDlg->Update(step*100, line);
			step++;
		} else if (line.StartsWith(wxT("Saving to"))) {
			progDlg->Update(step*100, line);
			step++;
		} else if (pattern.Matches(line)) {
			long percent = 0;
			if (pattern.GetMatch(line, 1).ToLong(&percent)) {
				progDlg->Update((step-1)*100 + percent);
			}
		}
    }

private:
	wxProgressDialog* progDlg;
	vector<MediaFile*>& files;
	bool terminated;
	int step;
	wxRegEx pattern;
	wxString lineOut;
	wxString lineErr;
};

//(*IdInit(MainWin)
const long MainWin::ID_MEDIA_LISTBOX = wxNewId();
const long MainWin::ID_TOOLBAR = wxNewId();
const long MainWin::ID_STATUSBAR = wxNewId();
//*)
const long ADD_FILE_ID = wxNewId();
const long REMOVE_FILE_ID = wxNewId();
const long RUN_ID = wxNewId();
const long SETTINGS_ID = wxNewId();
const long ABOUT_ID = wxNewId();

BEGIN_EVENT_TABLE(MainWin,wxFrame)
	//(*EventTable(MainWin)
	//*)
END_EVENT_TABLE()

MainWin::MainWin() {
	//(*Initialize(MainWin)
	Create(0, wxID_ANY, _("MP4 Joiner"), wxDefaultPosition, wxDefaultSize, wxDEFAULT_FRAME_STYLE, _T("wxID_ANY"));
	SetFocus();
	mediaListBox = new MediaListBox(this, ID_MEDIA_LISTBOX, wxPoint(128,176), wxSize(300,400), 0, 0, 0, wxDefaultValidator, _T("ID_MEDIA_LISTBOX"));
	statusBar = new wxStatusBar(this, ID_STATUSBAR, 0, _T("ID_STATUSBAR"));
	int __wxStatusBarWidths_1[1] = { -10 };
	int __wxStatusBarStyles_1[1] = { wxSB_NORMAL };
	statusBar->SetFieldsCount(1,__wxStatusBarWidths_1);
	statusBar->SetStatusStyles(1,__wxStatusBarStyles_1);
	SetStatusBar(statusBar);
	Center();
	
	Connect(ID_MEDIA_LISTBOX,wxEVT_COMMAND_LISTBOX_SELECTED,(wxObjectEventFunction)&MainWin::OnMediaSelect);
	//*)

	toolbar = new wxToolBar(this, ID_TOOLBAR, wxDefaultPosition, wxDefaultSize, wxTB_TEXT|wxTB_FLAT);
	SetToolBar(toolbar);
	toolbar->AddTool(ADD_FILE_ID, _("Add Video"), wxBITMAP_FROM_MEMORY(add), wxNullBitmap, wxITEM_NORMAL,
			_("Add video file"), wxT(""));
	toolbar->AddTool(REMOVE_FILE_ID,  _("Remove"), wxBITMAP_FROM_MEMORY(remove), wxNullBitmap, wxITEM_NORMAL,
			_("Remove selected video"), wxT(""));
	toolbar->AddTool(RUN_ID,  _("Join"), wxBITMAP_FROM_MEMORY(run), wxNullBitmap, wxITEM_NORMAL,
			_("Select output file and start join"), wxT(""));
#if wxCHECK_VERSION(2, 9, 0)
	toolbar->AddStretchableSpace();
#endif
	toolbar->AddTool(SETTINGS_ID, _("Language"), wxBITMAP_FROM_MEMORY(preferences),
			wxNullBitmap, wxITEM_NORMAL, _("Select language"), wxT(""));
	toolbar->AddTool(ABOUT_ID, _("About"), wxArtProvider::GetBitmap(wxART_INFORMATION, wxART_TOOLBAR),
			wxNullBitmap, wxITEM_NORMAL, _("About the program"), wxT(""));
	toolbar->Realize();
	UpdateToolBar();

	Connect(ADD_FILE_ID, wxEVT_COMMAND_TOOL_CLICKED, (wxObjectEventFunction)&MainWin::OnAddFileBt);
	Connect(REMOVE_FILE_ID, wxEVT_COMMAND_TOOL_CLICKED, (wxObjectEventFunction)&MainWin::OnRemoveFileBt);
	Connect(RUN_ID, wxEVT_COMMAND_TOOL_CLICKED, (wxObjectEventFunction)&MainWin::OnRunBt);
	Connect(SETTINGS_ID, wxEVT_COMMAND_TOOL_CLICKED, (wxObjectEventFunction)&MainWin::OnSettingsBt);
	Connect(ABOUT_ID, wxEVT_COMMAND_TOOL_CLICKED, (wxObjectEventFunction)&MainWin::OnAboutBt);
	Connect(wxEVT_CLOSE_WINDOW, (wxObjectEventFunction)&MainWin::OnClose);
	
	// Restore frame size/position
	if (s_config.IsMainWinMaximized()) {
		Maximize(true);
	} else {
		wxRect rect = s_config.GetMainWinLocation();
		if (rect.width > 0 && rect.height > 0)
			SetSize(rect);
		else
			SetSize(wxSize(500,550));
	}
}

MainWin::~MainWin() {
	//(*Destroy(MainWin)
	//*)
}

void MainWin::UpdateToolBar() {
	toolbar->EnableTool(REMOVE_FILE_ID, mediaListBox->GetSelection() >= 0);
	toolbar->EnableTool(RUN_ID, files.size() > 1);
	long totalSize = 0;
	for (vector<MediaFile*>::iterator it = files.begin(); it != files.end(); it++)
		totalSize += (*it)->GetSize() / 1024;
	wxString sizeStr = _("Total size");
	if (totalSize > 1024)
		sizeStr += wxString::Format(wxT(": %0.1f "), (double) totalSize / 1024) + _("GB");
	else
		sizeStr += wxString::Format(wxT(": %ld "), totalSize) + _("MB");
	statusBar->SetStatusText(sizeStr, 0);
}

void MainWin::OnAddFileBt(wxCommandEvent& event) {
	wxFileDialog fileDlg(this, _("Choose a file"), wxT(""), wxT(""),
			wxString(_("MP4 Files")) + wxT(" (*.mp4)|*.mp4|")
			+ wxString(_("All Files")) + wxT(" (*.*)|*.*"), wxFD_OPEN | wxFD_MULTIPLE);
	fileDlg.SetDirectory(s_config.GetLastAddDir() + wxFILE_SEP_PATH);
	if (fileDlg.ShowModal() != wxID_OK)
		return;
	s_config.SetLastAddDir(fileDlg.GetDirectory());
	wxArrayString paths;
	fileDlg.GetPaths(paths);
	for (unsigned int i = 0; i < paths.GetCount(); i++)
		files.push_back(new MediaFile(paths[i]));
	mediaListBox->RefreshAll();
	UpdateToolBar();
}

void MainWin::OnRemoveFileBt(wxCommandEvent& event) {
	delete files[mediaListBox->GetSelection()];
	files.erase(files.begin() + mediaListBox->GetSelection());
	mediaListBox->RefreshAll();
	UpdateToolBar();
}

void MainWin::OnRunBt(wxCommandEvent& event) {
	wxFileDialog fileDlg(this, _("Choose a file to save"), wxT(""), _("Output.mp4"),
		wxString(_("MP4 Files")) + wxT(" (*.mp4)|*.mp4|")
		+ wxString(_("All Files")) + wxT(" (*.*)|*.*"), wxFD_SAVE | wxFD_OVERWRITE_PROMPT);
	if (fileDlg.ShowModal() != wxID_OK)
		return;
	wxString fileName = fileDlg.GetPath();
	if (wxFile::Exists(fileName) && !wxRemoveFile(fileName))
		return;

	wxString cmd = wxT("MP4Box");
#ifdef __WXMSW__
		cmd = wxGetAppPath() + cmd;
#endif
	for (vector<MediaFile*>::iterator it = files.begin(); it != files.end(); it++)
		cmd += wxT(" -cat \"") + (*it)->GetFileName() + wxT('"');
	cmd += wxT(" \"") + fileName + wxT('"');
	
	wxProgressDialog progDlg(_("MP4 Joiner"), _("Joning the files"), (files.size() + 1)*100, this,
			wxPD_APP_MODAL | wxPD_CAN_ABORT | wxPD_ELAPSED_TIME | wxPD_ESTIMATED_TIME);
	progDlg.Show();
	
	// Execute
	MP4Process proc(&progDlg, files);
	proc.Redirect();
	long pid = wxExecute(cmd, wxEXEC_ASYNC, &proc);
	if (!pid) {
		wxLogError(wxT("Execution of '%s' failed."), cmd.c_str());
		return;
	}
	while (!proc.IsTerminated() && progDlg.IsVisible()) {
#if wxCHECK_VERSION(2, 9, 0)
		if (progDlg.WasCancelled() && wxProcess::Exists(pid)) {
			wxProcess::Kill(pid, wxSIGTERM);
			wxMilliSleep(500);
			if (wxProcess::Exists(pid))
				wxProcess::Kill(pid, wxSIGKILL);
		} else
#endif
		if (!proc.HasInput())
			wxMilliSleep(100);
		wxYield();
	}
}

void MainWin::OnSettingsBt(wxCommandEvent& event) {
	int langId = wxLANGUAGE_ENGLISH;
	if (wxLocale::FindLanguageInfo(s_config.GetLanguageCode()))
		langId = wxLocale::FindLanguageInfo(s_config.GetLanguageCode())->Language;
	int lang = ChooseLanguage(langId);
	if (lang == wxLANGUAGE_UNKNOWN)
		return;
	wxString languageCode = wxLocale::GetLanguageInfo(lang)->CanonicalName;
	if (languageCode == s_config.GetLanguageCode())
		return;
	s_config.SetLanguageCode(languageCode);
	wxMessageBox(_("Language change will not take effect until MP4Joiner is restarted"),
			GetTitle(), wxOK|wxICON_INFORMATION, this);
}

wxString FixEmail(const wxString& str) {
	wxString result = str;
	result.Replace(wxT(" at "), wxT("@"));
	return result;
}

void MainWin::OnAboutBt(wxCommandEvent& event) {
	wxAboutDialogInfo info;
	info.SetName(wxT("MP4Joiner"));
	info.SetVersion(APP_VERSION);
	info.SetDescription(_("Simple tool to join multiply MP4 files into one."));
	info.SetCopyright(FixEmail(wxString("(C) 2012 Alex Thring <alex at thuering.biz>", wxConvISO8859_1)));
	info.SetWebSite(wxT("http://www.mp4joiner.org"));
#ifdef __WXMSW__
	info.SetIcon(wxICON(mp4joiner));
#else
	info.SetIcon(wxICON_FROM_MEMORY(mp4joiner));
#endif
	wxAboutBox(info);
	
}

void MainWin::OnMediaSelect(wxCommandEvent& event) {
	UpdateToolBar();
}

void MainWin::OnClose(wxCloseEvent &event) {
	if (!IsIconized())
		s_config.SetMainWinLocation(GetRect(), IsMaximized());
	Destroy();
}
